#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import os
import sys

os.environ.setdefault("DJANGO_SETTINGS_MODULE", 'tests.test_settings')


def runtests():
    import django
    from django.conf import settings
    from django.test.utils import get_runner

    if hasattr(django, 'setup'):
        django.setup()

    TestRunner = get_runner(settings)
    test_runner = TestRunner(verbosity=2, interactive=True)
    failures = test_runner.run_tests(['tests'])
    sys.exit(bool(failures))


if __name__ == "__main__":
    runtests()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

from django.apps import apps
from django.db.utils import IntegrityError


class HorizontalRouter(object):
    def _get_horizontal_group(self, model):
        if hasattr(model, '_get_horizontal_group'):
            return model._get_horizontal_group()

    def _get_horizontal_index(self, model, hints):
        horizontal_group = self._get_horizontal_group(model)
        if not horizontal_group:
            return

        horizontal_key = hints.get('horizontal_key', None)
        if not horizontal_key:
            instance = hints.get('instance', None)
            if instance and isinstance(instance, model):
                return instance._horizontal_database_index
            if instance:
                # From foreign field
                horizontal_key = instance.pk

        if not horizontal_key:
            raise IntegrityError("Missing 'horizontal_key'")
        return model._get_or_create_horizontal_index(horizontal_key)

    def db_for_read(self, model, **hints):
        horizontal_index = self._get_horizontal_index(model, hints)
        if not horizontal_index:
            return
        return model._get_horizontal_db_for_read(horizontal_index)

    def db_for_write(self, model, **hints):
        horizontal_index = self._get_horizontal_index(model, hints)
        if not horizontal_index:
            return
        return model._get_horizontal_db_for_write(horizontal_index)

    def allow_relation(self, obj1, obj2, **hints):
        horizontal_group_1 = self._get_horizontal_group(obj1._meta.model)
        horizontal_group_2 = self._get_horizontal_group(obj2._meta.model)
        if not horizontal_group_1 or not horizontal_group_2:
            return

        if horizontal_group_1 != horizontal_group_2:
            return

        if obj1._state.db == obj2._state.db:
            return True

        if obj1._horizontal_database_index == obj2._horizontal_database_index:
            return True

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if 'model' in hints:
            model = hints['model']
        elif model_name:
            model = apps.get_model(app_label, model_name)
        else:
            return

        horizontal_group = self._get_horizontal_group(model)
        if not horizontal_group:
            return

        if db in model._get_horizontal_config()['DATABASE_SET']:
            return True



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import django
from django.apps import apps
from django.core.exceptions import ImproperlyConfigured

from .settings import get_config


def get_metadata_model():
    try:
        if (1, 11) > django.VERSION:
            return apps.get_model(get_config()['METADATA_MODEL'])
        else:
            return apps.get_model(get_config()['METADATA_MODEL'], require_ready=False)

    except ValueError:
        raise ImproperlyConfigured("METADATA_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(
            "METADATA_MODEL refers to model '%s' that has not been installed"
            % get_config()['METADATA_MODEL']
        )



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

from django.db.models import Model
from django.db.models.query import QuerySet
from django.db.utils import ProgrammingError


class HorizontalQuerySet(QuerySet):
    def __init__(self, model=None, **kwargs):
        super(HorizontalQuerySet, self).__init__(model=model, **kwargs)
        self._horizontal_key = None

    @classmethod
    def _get_horizontal_key_from_lookup_value(cls, lookup_value):
        if not lookup_value:
            return
        if isinstance(lookup_value, Model):
            return lookup_value.pk
        return lookup_value

    def _set_horizontal_key_from_params(self, kwargs):
        if self._horizontal_key is not None:
            return

        key_field = self.model._meta.get_field(self.model._get_horizontal_key())
        lookup_value = kwargs.get(key_field.attname, None) or kwargs.get(key_field.name, None)
        self._horizontal_key = self._get_horizontal_key_from_lookup_value(lookup_value)

    def _create_object_from_params(self, lookup, params):
        self._set_horizontal_key_from_params(lookup)
        return super(HorizontalQuerySet, self)._create_object_from_params(lookup, params)

    def _extract_model_params(self, defaults, **kwargs):
        self._set_horizontal_key_from_params(kwargs)
        return super(HorizontalQuerySet, self)._extract_model_params(defaults, **kwargs)

    def _filter_or_exclude(self, negate, *args, **kwargs):
        self._set_horizontal_key_from_params(kwargs)
        return super(HorizontalQuerySet, self)._filter_or_exclude(negate, *args, **kwargs)

    def create(self, **kwargs):
        self._set_horizontal_key_from_params(kwargs)
        return super(HorizontalQuerySet, self).create(**kwargs)

    def _clone(self, **kwargs):
        clone = super(HorizontalQuerySet, self)._clone(**kwargs)
        clone._horizontal_key = self._horizontal_key
        return clone

    @property
    def db(self):
        if self._db:
            return self._db

        if self._horizontal_key is None:
            raise ProgrammingError("Missing horizontal key field's filter")

        self._add_hints(horizontal_key=self._horizontal_key)
        return super(HorizontalQuerySet, self).db



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import random
import uuid

from django.core import checks
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models import options
from django.utils.functional import cached_property

from .manager import HorizontalManager
from .settings import get_config
from .utils import get_metadata_model


options.DEFAULT_NAMES += (
    'horizontal_group',
    'horizontal_key',
)


class AbstractHorizontalMetadata(models.Model):
    group = models.CharField(max_length=15)
    key = models.CharField(max_length=32)
    index = models.PositiveSmallIntegerField()

    class Meta(object):
        abstract = True
        unique_together = (
            ('group', 'key'),
        )


class AbstractHorizontalModel(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    objects = HorizontalManager()

    @classmethod
    def check(cls, **kwargs):
        errors = super(AbstractHorizontalModel, cls).check(**kwargs)
        clash_errors = cls._check_horizontal_meta(**kwargs)
        if clash_errors:
            errors.extend(clash_errors)
            return errors

        errors.extend(cls._check_horizontal_key(**kwargs))
        return errors

    @classmethod
    def _check_horizontal_meta(cls, **kwargs):
        errors = []
        if not cls._get_horizontal_group():
            errors.append(
                checks.Error(
                    "'horizontal_group' not configured.",
                    obj=cls,
                    id='horizon.E001',
                )
            )
        if not cls._get_horizontal_key():
            errors.append(
                checks.Error(
                    "'horizontal_key' not configured.",
                    obj=cls,
                    id='horizon.E001',
                )
            )
        return errors

    @classmethod
    def _check_horizontal_group(cls, **kwargs):
        if cls._get_horizontal_group() in get_config()['GROUPS']:
            return []
        return [
            checks.Error(
                "'horizontal_group' '%s' does not defined in settings."
                % cls._get_horizontal_group(),
                obj=cls,
                id='horizon.E002',
            ),
        ]

    @classmethod
    def _check_horizontal_key(cls, **kwargs):
        try:
            cls._meta.get_field(cls._get_horizontal_key())
            return []
        except FieldDoesNotExist:
            return [
                checks.Error(
                    "'horizontal_key' refers to the non-existent field '%s'."
                    % cls._get_horizontal_key(),
                    obj=cls,
                    id='horizon.E003',
                ),
            ]

    def _get_unique_checks(self, exclude=None):
        if exclude is None:
            exclude = []

        fields_with_class = [(self.__class__, self._meta.local_fields)]
        for parent_class in self._meta.get_parent_list():
            fields_with_class.append((parent_class, parent_class._meta.local_fields))

        # Exclude default field's unique check and add field's unique check with horizontal key
        unique_fields_for_exclude = []
        unique_fields_with_horizontal_key = []
        for model_class, fields in fields_with_class:
            for f in fields:
                name = f.name
                if name in exclude:
                    continue
                if f.unique:
                    unique_fields_for_exclude.append(name)
                    unique_fields_with_horizontal_key.append(
                        (model_class, (self._get_horizontal_key(), name)),
                    )

        unique_checks, date_checks = super(AbstractHorizontalModel, self)._get_unique_checks(
            exclude=exclude + unique_fields_for_exclude,
        )
        return unique_checks + unique_fields_with_horizontal_key, date_checks

    @classmethod
    def _get_horizontal_config(cls):
        return get_config()['GROUPS'][cls._get_horizontal_group()]

    @classmethod
    def _get_horizontal_group(cls):
        horizontal_group = getattr(cls._meta, 'horizontal_group', None)
        if horizontal_group:
            return horizontal_group

        for parent_class in cls._meta.get_parent_list():
            horizontal_group = getattr(parent_class._meta, 'horizontal_group', None)
            if horizontal_group:
                return horizontal_group

    @classmethod
    def _get_horizontal_key(cls):
        horizontal_key = getattr(cls._meta, 'horizontal_key', None)
        if horizontal_key:
            return horizontal_key

        for parent_class in cls._meta.get_parent_list():
            horizontal_key = getattr(parent_class._meta, 'horizontal_key', None)
            if horizontal_key:
                return horizontal_key

    @classmethod
    def _get_horizontal_db_for_read(cls, index):
        return random.choice(cls._get_horizontal_config()['DATABASES'][index]['read'])

    @classmethod
    def _get_horizontal_db_for_write(cls, index):
        return cls._get_horizontal_config()['DATABASES'][index]['write']

    @classmethod
    def _get_or_create_horizontal_index(cls, horizontal_key):
        metadata_model = get_metadata_model()
        metadata, created = metadata_model.objects.get_or_create(
            group=cls._get_horizontal_group(),
            key=horizontal_key,
            defaults={
                'index': random.choice(cls._get_horizontal_config()['PICKABLES'])
            },
        )
        return metadata.index

    @cached_property
    def _horizontal_key(self):
        key_field = self._meta.get_field(self._get_horizontal_key())
        return getattr(self, key_field.attname)

    @cached_property
    def _horizontal_database_index(self):
        return self._get_or_create_horizontal_index(self._horizontal_key)

    class Meta(object):
        abstract = True
        horizontal_group = None
        horizontal_key = None



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

from django.db.models.manager import Manager

from .query import HorizontalQuerySet


class HorizontalManager(Manager.from_queryset(HorizontalQuerySet)):
    use_for_related_fields = True
    use_in_migrations = True
    silence_use_for_related_fields_deprecation = True  # For Django<1.10

    def __init__(self):
        super(HorizontalManager, self).__init__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

import horizon

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Django>=1.8',
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='django-horizon',
    version=horizon.__version__,
    description=(
        "Simple database sharding (horizontal partitioning) library for Django applications."
    ),
    long_description=readme + '\n\n' + history,
    author=horizon.__author__,
    author_email=horizon.__email__,
    url='https://github.com/uncovertruth/django-horizon',
    packages=find_packages(exclude=('tests', 'docs')),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='django-horizon, django, sharding, horizontal partitioning, database',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',  # Allow failure because SQLite is too old
        'Programming Language :: Python :: 2.7',  # Allow failure because SQLite is too old
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='runtests.runtests',
    tests_require=test_requirements
)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.utils.lru_cache import lru_cache


CONFIG_DEFAULTS = {
    'GROUPS': {},
    'METADATA_MODEL': None,
}


@lru_cache()
def get_config():
    USER_CONFIG = getattr(settings, 'HORIZONTAL_CONFIG', {})

    CONFIG = CONFIG_DEFAULTS.copy()
    CONFIG.update(USER_CONFIG)

    for name, horizontal_group in CONFIG['GROUPS'].items():
        horizontal_group['DATABASE_SET'] = set()
        for key, member in horizontal_group['DATABASES'].items():
            horizontal_group['DATABASE_SET'].add(member['write'])
            horizontal_group['DATABASE_SET'].update(member.get('read', []))

            if 'read' not in member:
                member['read'] = [member['write']]

        if 'PICKABLES' not in horizontal_group:
            horizontal_group['PICKABLES'] = [int(i) for i in horizontal_group['DATABASES'].keys()]

    return CONFIG
